"""Module containing SOI-module 'Autentiseringstavle'."""
import string
from PySide2.QtWidgets import QTableWidget, QTableWidgetItem
from PySide2 import QtGui
from PySide2.QtCore import Qt
from soitool.coder import get_code_set, get_code
from soitool.modules.module_base import (
    ModuleBase,
    get_table_size,
    resize_table,
    HEADLINE_FONT,
)

START_NO_OF_AUTHENTICATION_CODES = 10
CODE_LENGTH = 25

# Has to be 'ascii', 'digits' or 'combo'. Codes will consist of A-Z if
# 'ascii', 0-9 if 'digits' and A-Z+0-9 if 'combo'.
CODE_CHARACTERS = "ascii"

# Characters for first column. It's length determines maximum number of
# codes (rows).
ROW_IDENTIFIERS = string.ascii_uppercase

# Adds space between sets of characters, 0 => no spaces
# If code is 123456 and interval is 2, code will be 12 34 56
SPACE_INTERVAL = 5

HEADLINE_TEXT = "Autentiseringstavle"


class Meta(type(ModuleBase), type(QTableWidget)):
    """Used as a metaclass to enable multiple inheritance."""


class AuthenticationBoardModule(ModuleBase, QTableWidget, metaclass=Meta):
    """Modified QTableWidget representing a 'Autentiseringstavle'.

    By default, the widget initializes with a headline, a row-count of
    START_NO_OF_AUTHENTICATION_CODES and three columns.
    Row x in the first column contains the character ROW_IDENTIFIERS[x].
    Row x in the second column contains the character x.
    Row x in the third column contains an authentication code.

    If parameters are given, the widget initializes accordingly:
    'size' is a dict: {"width": int, "height": int},
    'data' is a 2D list where data[0] is the headline,
    and data[x][y] represents row x, column y.

    The widget does not use more room than needed, and resizes dynamically.
    It has shortcuts for adding and removing rows.

    Inherits from ModuleBase and QTableWidget.
    ModuleBase is used as an interface, it's methods are overridden.
    """

    def __init__(self, size=None, data=None):
        self.type = "AuthenticationBoardModule"
        QTableWidget.__init__(self)
        ModuleBase.__init__(self)

        if CODE_CHARACTERS == "ascii":
            self.code_characters = string.ascii_uppercase
        elif CODE_CHARACTERS == "digits":
            self.code_characters = string.digits
        elif CODE_CHARACTERS == "combo":
            self.code_characters = string.ascii_uppercase + string.digits
        else:
            raise ValueError(
                "Invalid value for CONSTANT 'CODE_CHARACTERS': "
                "'{}'".format(CODE_CHARACTERS)
            )

        # Remove headers and scrollbars
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # If parameters are None, generate new table
        if size is None and data is None:
            self.generate_table()
            self.resizeColumnsToContents()
            self.insert_headline()

            # Resize height of rows and set size of window
            resize_table(self, resize_column=False)
        else:
            self.setColumnCount(len(data[1]))
            self.setRowCount(len(data) - 1)  # - 1 to skip headline

            # Set cell-items
            for i in range(self.rowCount()):
                for j in range(self.columnCount()):
                    item = QTableWidgetItem(data[i + 1][j])  # +1 skip headline
                    self.setItem(i, j, item)

            self.resizeColumnsToContents()
            resize_table(self, resize_column=False)
            self.setFixedWidth(size["width"])
            self.setFixedHeight(size["height"])

            self.insert_headline(data[0])

    def generate_table(self):
        """Insert row identifiers and authentication codes."""
        # Set number of rows and columns
        self.setRowCount(START_NO_OF_AUTHENTICATION_CODES)
        self.setColumnCount(3)

        # Generate codes
        codes = list(
            get_code_set(
                START_NO_OF_AUTHENTICATION_CODES,
                CODE_LENGTH,
                CODE_CHARACTERS,
                SPACE_INTERVAL,
            )
        )

        # Insert table data
        for i in range(START_NO_OF_AUTHENTICATION_CODES):
            # Insert non-editable row identifier in first column
            item_first = QTableWidgetItem(ROW_IDENTIFIERS[i])
            item_first.setFlags(item_first.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 0, item_first)

            # Insert non-editable row identifier (int) in second column
            item_second = QTableWidgetItem(str(i))
            item_second.setFlags(item_second.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 1, item_second)

            # Insert non-editable code in third column
            item_third = QTableWidgetItem(codes[i])
            item_third.setFlags(item_third.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 2, item_third)

    def insert_headline(self, text=HEADLINE_TEXT):
        """Insert headline text.

        Parameters
        ----------
        text : string, optional
            The headline text, by default HEADLINE_TEXT
        """
        item_headline = QTableWidgetItem(text)
        item_headline.setTextAlignment(Qt.AlignHCenter)
        item_headline.setFont(HEADLINE_FONT)

        self.insertRow(0)
        self.setItem(0, 0, item_headline)
        self.setSpan(0, 0, 1, self.columnCount())  # Make cell span all columns

    def generate_unique_authentication_code(self):
        """Generate authentication-code that does not already exist.

        Returns
        -------
        string
            Generated, unique authentication-code
        """
        # Get existing codes
        existing_codes = self.get_codes()

        # Randomly generate a new code until it is unique
        unique_code = False
        while not unique_code:
            code = get_code(CODE_LENGTH, CODE_CHARACTERS, SPACE_INTERVAL)
            unique_code = code not in existing_codes

        return code

    def keyPressEvent(self, event):
        """Add or remove row when 'Ctrl + +' and 'Ctrl + -' are pressed."""
        if (
            event.modifiers() == Qt.ControlModifier
            and event.key() == Qt.Key_Plus
        ):
            self.add_row()
        elif (
            event.modifiers() == Qt.ControlModifier
            and event.key() == Qt.Key_Underscore
        ):
            self.remove_row()
        else:
            super(AuthenticationBoardModule, self).keyPressEvent(event)

    def add_row(self):
        """Insert row below the selected row and add data."""
        row_index = self.currentRow()
        # If maximum amount of rows not reached and a row is selected
        # (+ 1 to skip row containing headline)
        if self.rowCount() < len(ROW_IDENTIFIERS) + 1 and row_index != -1:
            # Generate unique code and insert row
            code = self.generate_unique_authentication_code()
            self.insertRow(row_index + 1)

            # Loop through all rows starting with the new row
            for i in range(row_index + 1, self.rowCount()):
                # Insert row identifier in first column
                item_first = QTableWidgetItem(self.code_characters[i - 1])
                item_first.setFlags(item_first.flags() ^ Qt.ItemIsEditable)
                self.setItem(i, 0, item_first)

                # Insert row identifier (int) in second column
                item_second = QTableWidgetItem(str(i - 1))
                item_second.setFlags(item_second.flags() ^ Qt.ItemIsEditable)
                self.setItem(i, 1, item_second)

            # Insert authentication-code in third column
            item_third = QTableWidgetItem(code)
            item_third.setFlags(item_third.flags() ^ Qt.ItemIsEditable)
            self.setItem(row_index + 1, 2, item_third)

            # Resize code-column in case it got wider
            # Example: 'BGD' is wider than 'III' (depending on font)
            self.resizeColumnToContents(2)
            resize_table(self, resize_column=False)

    def remove_row(self):
        """Remove selected row."""
        row_index = self.currentRow()
        # If at least one row (+ headline-row) exists and a row other than
        # headline-row is selected
        if self.rowCount() > 2 and row_index != 0 and row_index != -1:
            # Remove row
            self.removeRow(row_index)
            # 'Decrease' row identifiers below the removed row
            # If first row is removed, identifier A,B,C becomes A,B (not B,C)
            for i in range(row_index, self.rowCount()):
                self.item(i, 0).setText(self.code_characters[i - 1])
                self.item(i, 1).setText(str(i - 1))
        resize_table(self, resize_column=False)

    def get_codes(self):
        """Get all authentication-codes in table.

        Returns
        -------
        List
            List containing authentication-codes.
        """
        codes = []

        # Start with row 1 to skip headline-row
        for i in range(1, self.rowCount()):
            codes.append(self.item(i, 2).text())

        return codes

    def get_size(self):
        """Return size of widget."""
        return get_table_size(self)

    def get_data(self):
        """Return list containing all data.

        Returns
        -------
        List
            List[0] contains headline,
            list[x][y] represents value of row x, column y.
        """
        content = []
        item_headline = self.item(0, 0)
        if item_headline is not None:
            content.append(item_headline.text())

        for i in range(1, self.rowCount()):
            row = []
            for j in range(self.columnCount()):
                row.append(self.item(i, j).text())
            content.append(row)

        return content

    @staticmethod
    def get_user_friendly_name():
        """Get user-friendly name of module."""
        return "Autentiseringstavle"

    @staticmethod
    def get_icon():
        """Get icon of module."""
        return QtGui.QIcon("soitool/media/authenticationboardmodule.png")
